package dev.shadowsoffire.apotheosis.adventure.boss;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import com.google.common.base.Preconditions;
import com.mojang.datafixers.kinds.App;
import com.mojang.datafixers.kinds.Applicative;
import com.mojang.datafixers.util.Function12;
import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import dev.shadowsoffire.apotheosis.Apotheosis;
import dev.shadowsoffire.apotheosis.adventure.AdventureConfig;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixHelper;
import dev.shadowsoffire.apotheosis.adventure.boss.ApothBoss;
import dev.shadowsoffire.apotheosis.adventure.boss.BossStats;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import dev.shadowsoffire.apotheosis.adventure.loot.LootController;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityClamp;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry;
import dev.shadowsoffire.apotheosis.ench.asm.EnchHooks;
import dev.shadowsoffire.apotheosis.util.NameHelper;
import dev.shadowsoffire.apotheosis.util.SupportingEntity;
import dev.shadowsoffire.placebo.codec.CodecProvider;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import dev.shadowsoffire.placebo.json.ChancedEffectInstance;
import dev.shadowsoffire.placebo.json.GearSet;
import dev.shadowsoffire.placebo.json.GearSetRegistry;
import dev.shadowsoffire.placebo.json.NBTAdapter;
import dev.shadowsoffire.placebo.json.RandomAttributeModifier;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.ComponentContents;
import net.minecraft.network.chat.MutableComponent;
import net.minecraft.network.chat.Style;
import net.minecraft.network.chat.contents.TranslatableContents;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.util.Mth;
import net.minecraft.util.RandomSource;
import net.minecraft.world.effect.MobEffectInstance;
import net.minecraft.world.effect.MobEffects;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.EntityType;
import net.minecraft.world.entity.EquipmentSlot;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.Mob;
import net.minecraft.world.entity.ai.goal.Goal;
import net.minecraft.world.entity.ai.goal.target.NearestAttackableTargetGoal;
import net.minecraft.world.entity.npc.Villager;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.enchantment.Enchantment;
import net.minecraft.world.item.enchantment.EnchantmentHelper;
import net.minecraft.world.item.enchantment.EnchantmentInstance;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.ServerLevelAccessor;
import net.minecraft.world.phys.AABB;
import net.minecraftforge.registries.ForgeRegistries;
import org.jetbrains.annotations.Nullable;

public final class ApothBoss implements CodecProvider<ApothBoss>, WeightedDynamicRegistry.ILuckyWeighted, WeightedDynamicRegistry.IDimensional, RarityClamp, GameStagesCompat.IStaged {
  public static final Codec<AABB> AABB_CODEC;
  
  public static final Codec<ApothBoss> CODEC;
  
  public static final Predicate<Goal> IS_VILLAGER_ATTACK;
  
  protected final int weight;
  
  protected final float quality;
  
  protected final EntityType<?> entity;
  
  protected final AABB size;
  
  protected final Map<LootRarity, BossStats> stats;
  
  @Nullable
  protected final Set<String> stages;
  
  protected final List<GearSet.SetPredicate> gearSets;
  
  @Nullable
  protected final CompoundTag nbt;
  
  protected final Set<ResourceLocation> dimensions;
  
  protected final LootRarity minRarity;
  
  protected final LootRarity maxRarity;
  
  @Nullable
  protected final SupportingEntity mount;
  
  static {
    AABB_CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.DOUBLE.fieldOf("width").forGetter(()), (App)Codec.DOUBLE.fieldOf("height").forGetter(())).apply((Applicative)inst, ()));
    CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.intRange(0, 2147483647).fieldOf("weight").forGetter(WeightedDynamicRegistry.ILuckyWeighted::getWeight), (App)PlaceboCodecs.nullableField(Codec.floatRange(0.0F, Float.MAX_VALUE), "quality", Float.valueOf(0.0F)).forGetter(WeightedDynamicRegistry.ILuckyWeighted::getQuality), (App)ForgeRegistries.ENTITY_TYPES.getCodec().fieldOf("entity").forGetter(()), (App)AABB_CODEC.fieldOf("size").forGetter(()), (App)LootRarity.mapCodec(BossStats.CODEC).fieldOf("stats").forGetter(()), (App)PlaceboCodecs.nullableField(PlaceboCodecs.setOf((Codec)Codec.STRING), "stages").forGetter(()), (App)GearSet.SetPredicate.CODEC.listOf().fieldOf("valid_gear_sets").forGetter(()), (App)PlaceboCodecs.nullableField(NBTAdapter.EITHER_CODEC, "nbt").forGetter(()), (App)PlaceboCodecs.setOf(ResourceLocation.f_135803_).fieldOf("dimensions").forGetter(()), (App)LootRarity.CODEC.fieldOf("min_rarity").forGetter(()), (App)LootRarity.CODEC.fieldOf("max_rarity").forGetter(()), (App)PlaceboCodecs.nullableField(SupportingEntity.CODEC, "mount").forGetter(())).apply((Applicative)inst, ApothBoss::new));
    IS_VILLAGER_ATTACK = (a -> (a instanceof NearestAttackableTargetGoal && ((NearestAttackableTargetGoal)a).f_26048_ == Villager.class));
  }
  
  public ApothBoss(int weight, float quality, EntityType<?> entity, AABB size, Map<LootRarity, BossStats> stats, Optional<Set<String>> stages, List<GearSet.SetPredicate> armorSets, Optional<CompoundTag> nbt, Set<ResourceLocation> dimensions, LootRarity minRarity, LootRarity maxRarity, Optional<SupportingEntity> mount) {
    this.weight = weight;
    this.quality = quality;
    this.entity = entity;
    this.size = size;
    this.stats = stats;
    this.stages = stages.orElse(null);
    this.gearSets = armorSets;
    this.nbt = nbt.orElse(null);
    this.dimensions = dimensions;
    this.minRarity = minRarity;
    this.maxRarity = maxRarity;
    this.mount = mount.orElse(null);
    Preconditions.checkArgument((minRarity.ordinal() <= maxRarity.ordinal()), "Min rarity must be less than or equal to max rarity.");
  }
  
  public int getWeight() {
    return this.weight;
  }
  
  public float getQuality() {
    return this.quality;
  }
  
  public LootRarity getMinRarity() {
    return this.minRarity;
  }
  
  public LootRarity getMaxRarity() {
    return this.maxRarity;
  }
  
  public AABB getSize() {
    return this.size;
  }
  
  public EntityType<?> getEntity() {
    return this.entity;
  }
  
  public Mob createBoss(ServerLevelAccessor world, BlockPos pos, RandomSource random, float luck) {
    return createBoss(world, pos, random, luck, null);
  }
  
  public Mob createBoss(ServerLevelAccessor world, BlockPos pos, RandomSource random, float luck, @Nullable LootRarity rarity) {
    CompoundTag fakeNbt = (this.nbt == null) ? new CompoundTag() : this.nbt;
    fakeNbt.m_128359_("id", EntityType.m_20613_(this.entity).toString());
    Mob entity = (Mob)EntityType.m_20645_(fakeNbt, (Level)world.m_6018_(), Function.identity());
    if (this.nbt != null)
      entity.m_20258_(this.nbt); 
    initBoss(random, entity, luck, rarity);
    if (this.nbt != null)
      entity.m_7378_(this.nbt); 
    if (this.mount != null) {
      Mob mountedEntity = this.mount.create((Level)world.m_6018_(), pos.m_123341_() + 0.5D, pos.m_123342_(), pos.m_123343_() + 0.5D);
      entity.m_7998_((Entity)mountedEntity, true);
      entity = mountedEntity;
    } 
    entity.m_7678_(pos.m_123341_() + 0.5D, pos.m_123342_(), pos.m_123343_() + 0.5D, random.m_188501_() * 360.0F, 0.0F);
    return entity;
  }
  
  public void initBoss(RandomSource rand, Mob entity, float luck, @Nullable LootRarity rarity) {
    if (rarity == null)
      rarity = LootRarity.random(rand, luck, this); 
    rarity = clamp(rarity);
    BossStats stats = this.stats.get(rarity);
    int duration = (entity instanceof net.minecraft.world.entity.monster.Creeper) ? 6000 : Integer.MAX_VALUE;
    for (ChancedEffectInstance inst : stats.effects()) {
      if (rand.m_188501_() <= inst.chance())
        entity.m_7292_(inst.create(rand, duration)); 
    } 
    for (RandomAttributeModifier modif : stats.modifiers())
      modif.apply(rand, (LivingEntity)entity); 
    entity.f_21345_.f_25345_.removeIf(IS_VILLAGER_ATTACK);
    String name = NameHelper.setEntityName(rand, entity);
    GearSet set = GearSetRegistry.INSTANCE.getRandomSet(rand, luck, this.gearSets);
    set.apply((LivingEntity)entity);
    boolean anyValid = false;
    for (EquipmentSlot t : EquipmentSlot.values()) {
      ItemStack s = entity.m_6844_(t);
      if (!s.m_41619_() && !LootCategory.forItem(s).isNone()) {
        anyValid = true;
        break;
      } 
    } 
    if (!anyValid)
      throw new RuntimeException("Attempted to apply boss gear set " + GearSetRegistry.INSTANCE.getKey(set) + " but it had no valid affix loot items generated."); 
    int guaranteed = rand.m_188503_(6);
    ItemStack temp = entity.m_6844_(EquipmentSlot.values()[guaranteed]);
    while (temp.m_41619_() || LootCategory.forItem(temp) == LootCategory.NONE) {
      guaranteed = rand.m_188503_(6);
      temp = entity.m_6844_(EquipmentSlot.values()[guaranteed]);
    } 
    for (EquipmentSlot s : EquipmentSlot.values()) {
      ItemStack stack = entity.m_6844_(s);
      if (!stack.m_41619_()) {
        if (s.ordinal() == guaranteed)
          entity.m_21409_(s, 2.0F); 
        if (s.ordinal() == guaranteed) {
          entity.m_8061_(s, modifyBossItem(stack, rand, name, luck, rarity, stats));
          entity.m_6593_((Component)((MutableComponent)entity.m_7770_()).m_130948_(Style.f_131099_.m_131148_(rarity.getColor())));
        } else if (rand.m_188501_() < stats.enchantChance()) {
          enchantBossItem(rand, stack, Apotheosis.enableEnch ? stats.enchLevels()[0] : stats.enchLevels()[1], true);
          entity.m_8061_(s, stack);
        } 
      } 
    } 
    entity.getPersistentData().m_128379_("apoth.boss", true);
    entity.getPersistentData().m_128359_("apoth.rarity", RarityRegistry.INSTANCE.getKey((CodecProvider)rarity).toString());
    entity.m_21153_(entity.m_21233_());
    if (AdventureConfig.bossGlowOnSpawn)
      entity.m_7292_(new MobEffectInstance(MobEffects.f_19619_, 3600)); 
  }
  
  public static void enchantBossItem(RandomSource rand, ItemStack stack, int level, boolean treasure) {
    List<EnchantmentInstance> ench = EnchantmentHelper.m_220297_(rand, stack, level, treasure);
    Map<Enchantment, Integer> map = (Map<Enchantment, Integer>)ench.stream().filter(d -> !d.f_44947_.m_6589_()).collect(Collectors.toMap(d -> d.f_44947_, d -> Integer.valueOf(d.f_44948_), Math::max));
    map.putAll(EnchantmentHelper.m_44831_(stack));
    EnchantmentHelper.m_44865_(map, stack);
  }
  
  public static ItemStack modifyBossItem(ItemStack stack, RandomSource rand, String bossName, float luck, LootRarity rarity, BossStats stats) {
    enchantBossItem(rand, stack, Apotheosis.enableEnch ? stats.enchLevels()[2] : stats.enchLevels()[3], true);
    NameHelper.setItemName(rand, stack);
    stack = LootController.createLootItem(stack, LootCategory.forItem(stack), rarity, rand);
    String bossOwnerName = String.format(NameHelper.ownershipFormat, new Object[] { bossName });
    Component name = AffixHelper.getName(stack);
    if (!bossName.isEmpty()) {
      ComponentContents componentContents = name.m_214077_();
      if (componentContents instanceof TranslatableContents) {
        TranslatableContents tc = (TranslatableContents)componentContents;
        String oldKey = tc.m_237508_();
        String newKey = "misc.apotheosis.affix_name.two".equals(oldKey) ? "misc.apotheosis.affix_name.three" : "misc.apotheosis.affix_name.four";
        Object[] newArgs = new Object[(tc.m_237523_()).length + 1];
        newArgs[0] = bossOwnerName;
        for (int i = 1; i < newArgs.length; i++)
          newArgs[i] = tc.m_237523_()[i - 1]; 
        MutableComponent mutableComponent = Component.m_237110_(newKey, newArgs).m_130948_(name.m_7383_().m_131155_(Boolean.valueOf(false)));
        AffixHelper.setName(stack, (Component)mutableComponent);
      } 
    } 
    Map<Enchantment, Integer> enchMap = new HashMap<>();
    for (Map.Entry<Enchantment, Integer> e : (Iterable<Map.Entry<Enchantment, Integer>>)EnchantmentHelper.m_44831_(stack).entrySet()) {
      if (e.getKey() != null)
        enchMap.put(e.getKey(), Integer.valueOf(Math.min(EnchHooks.getMaxLevel(e.getKey()), ((Integer)e.getValue()).intValue() + rand.m_188503_(2)))); 
    } 
    if (AdventureConfig.curseBossItems) {
      ItemStack stk = stack;
      List<Enchantment> curses = (List<Enchantment>)ForgeRegistries.ENCHANTMENTS.getValues().stream().filter(e -> (e.canApplyAtEnchantingTable(stk) && e.m_6589_())).collect(Collectors.toList());
      if (!curses.isEmpty()) {
        Enchantment curse = curses.get(rand.m_188503_(curses.size()));
        enchMap.put(curse, Integer.valueOf(Mth.m_216271_(rand, 1, EnchHooks.getMaxLevel(curse))));
      } 
    } 
    EnchantmentHelper.m_44865_(enchMap, stack);
    stack.m_41783_().m_128379_("apoth_boss", true);
    return stack;
  }
  
  public ApothBoss validate(ResourceLocation key) {
    Preconditions.checkArgument((this.weight >= 0), "Boss Item " + key + " has a negative weight!");
    Preconditions.checkArgument((this.quality >= 0.0F), "Boss Item " + key + " has a negative quality!");
    Preconditions.checkNotNull(this.entity, "Boss Item " + key + " has null entity type!");
    Preconditions.checkNotNull(this.size, "Boss Item " + key + " has no size!");
    if (this.minRarity != null)
      Preconditions.checkArgument((this.maxRarity == null || this.maxRarity.isAtLeast(this.minRarity))); 
    if (this.maxRarity != null)
      Preconditions.checkArgument((this.minRarity == null || this.maxRarity.isAtLeast(this.minRarity))); 
    if (this.mount != null)
      Preconditions.checkNotNull(this.mount.entity, "Boss Item " + key + " has an invalid mount"); 
    LootRarity r = this.minRarity;
    while (r != this.maxRarity) {
      Preconditions.checkNotNull(this.stats.get(r));
      r = r.next();
    } 
    return this;
  }
  
  public Set<ResourceLocation> getDimensions() {
    return this.dimensions;
  }
  
  public Set<String> getStages() {
    return this.stages;
  }
  
  public Codec<? extends ApothBoss> getCodec() {
    return CODEC;
  }
}
